import asyncio
import logging
import json
import os

from decimal import Decimal
from itertools import product
import requests

from pscr_common.aws import get_boto3_client
from pscr_common.metrics import PrometheusRegistry

from index_registry.dynamodb.stats_table import StatsTable
from index_registry.maintenance.metrics import Metrics
from index_registry.utils.config import Config

_logger = logging.getLogger(__name__)


# The same function is used to calculate the average in the searchers
def divide_and_round_up(a, b):
    if b != 0:
        return int((a + b - 1) / b)
    else:
        return 1


class StatsEngine(object):
    def __init__(self, config: Config):
        self.config = config
        self.stats_table = StatsTable(config)
        self.autoscaling_api = get_boto3_client("autoscaling")
        self._registry = PrometheusRegistry()

    def gather_stats_table_data(self):
        table_scan_data = self.stats_table.scan_table()

        # Generate all the possible keys in a db record and save them in a list
        all_keys = []
        for key in product(
            ["default_", "longTerm_"], ["merging_", "optimized_"], ["count", "size"]
        ):
            all_keys.append("".join(key))

        # All org statistics will be stored in a dic with their org id as a key and stats as a value
        org_stats = {}

        for item in table_scan_data:
            org_data = item
            # Remove the org_id as it is unnecessary because it will be the key in the dic
            org_id = org_data.pop("org_id")

            # Fill any missing keys
            for key in all_keys:
                if key not in item:
                    org_data[key] = 0

            skip = True
            # Check if all fields of that org are empty, if so don't include it in the stats file
            for key in all_keys:
                if org_data[key] != 0:
                    skip = False
                    break

            if not skip:
                org_stats[org_id] = org_data

        return org_stats

    def get_searchers_jvms_count_ecs(self):
        # Get the searcher autoscaling group names
        auto_scaling_group_names = self.config.get(
            "stats", "searcher_autoscaling_groups"
        ).split(", ")

        searchers_desired_capacity = 0

        try:
            # Make a call for the autoscaling groups
            auto_scale_groups = self.autoscaling_api.describe_auto_scaling_groups(
                AutoScalingGroupNames=auto_scaling_group_names
            )["AutoScalingGroups"]

            # Find the two searchers scaling groups and get the desired capacity
            for group in auto_scale_groups:
                searchers_desired_capacity += group["DesiredCapacity"]

            # We multiply by 2 because each instance is hosting two searcher tasks: A and B
            # and we are interested in the count of JVMs
            return searchers_desired_capacity * 2
        except Exception as error:
            _logger.error("Error when making the describe autoscaling groups call")
            raise Exception(repr(error))

    def create_stats(self):
        # Get the necessary data to build the statistics
        orgs_stats = self.gather_stats_table_data()

        searchers_desired_capacity = -1

        # Support for both vdp and commercial
        if os.environ.get("STATS_VDP_TIMER") is not None:
            searchers_desired_capacity = self.get_searchers_jvms_count_k8s()
        else:
            searchers_desired_capacity = self.get_searchers_jvms_count_ecs()

        stats = {
            "searchers_desired_capacity": searchers_desired_capacity,
            "stats": orgs_stats,
        }

        return stats

    def report_org_metrics(self, stats, config):
        # Report orgs averages to wavefront if reporting is enabled
        if config.get("stats", "reporting_enabled").lower() != "true":
            return

        reporter = Metrics.get_reporter(self.config, self._registry)
        if reporter is None:
            _logger.error("Something went wrong with creating the reporter")
            raise EnvironmentError("Couldn't find the env vars for creating reporter")

        try:
            # Calculate and report the average values for each org
            for org, orgStats in stats["stats"].items():
                optimized_average = divide_and_round_up(
                    (
                        int(orgStats["longTerm_optimized_count"])
                        + int(orgStats["default_optimized_count"])
                    ),
                    int(stats["searchers_desired_capacity"]),
                )

                merging_average = divide_and_round_up(
                    (
                        int(orgStats["longTerm_merging_count"])
                        + int(orgStats["default_merging_count"])
                    ),
                    int(stats["searchers_desired_capacity"]),
                )

                self._registry.gauge(
                    "hosted_indexes.average_optimized", extra_labels={"org_key": org}
                ).set_value(optimized_average)
                self._registry.gauge(
                    "hosted_indexes.average_merging", extra_labels={"org_key": org}
                ).set_value(merging_average)
            self._registry.gauge("searchers_desired_count").set_value(
                stats["searchers_desired_capacity"]
            )
            reporter.report_now(self._registry, "index_registry")
            print("Reported data to wavefront successfully")

        except Exception:
            _logger.error("Stats lambda is not sending statistics to wavefront")
            raise Exception("Stats lambda is not sending statistics to wavefront")

    def get_searchers_jvms_count_k8s(self):
        sa_prefix = "/var/run/secrets/kubernetes.io/serviceaccount"
        ns = open(f"{sa_prefix}/namespace").read()
        token = open(f"{sa_prefix}/token").read()
        ca_cert_path = f"{sa_prefix}/ca.crt"
        host = os.environ.get("KUBERNETES_SERVICE_HOST", None)
        port = os.environ.get("KUBERNETES_SERVICE_PORT", None)
        # Typically that would be searcher-a and searcher-b
        searcher_daemonsets_names = os.environ.get(
            "STATS_SEARCHER_DAEMONSET_NAME", None
        ).split(",")

        searchers_jvms_count = 0

        for deamonset_name in searcher_daemonsets_names:
            r = requests.get(
                f"https://{host}:{port}/apis/apps/v1/namespaces/{ns}/daemonsets/{deamonset_name}",
                verify=ca_cert_path,
                headers={"Authorization": f"Bearer {token}"},
            )
            res = r.json()

            if res["kind"] == "DaemonSet":
                searchers_jvms_count += res["status"]["desiredNumberScheduled"]
            else:
                _logger.warning(
                    "The searchers daemonsets that we passed as ENV variable arent correct! Will loop "
                    "through all daemonsets in the namesapce!"
                )
                r = requests.get(
                    f"https://{host}:{port}/apis/apps/v1/namespaces/{ns}/daemonsets",
                    verify=ca_cert_path,
                    headers={"Authorization": f"Bearer {token}"},
                )
                res = r.json()

                for daemonset in res["items"]:
                    if any(
                        searcher in daemonset["metadata"]["name"]
                        for searcher in searcher_daemonsets_names
                    ):
                        searchers_jvms_count += daemonset["status"][
                            "desiredNumberScheduled"
                        ]

                break

        return searchers_jvms_count


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        return json.JSONEncoder.default(self, obj)
