from sqlalchemy import Column, DateTime, String, Integer, ForeignKey, ForeignKeyConstraint, Boolean, Index as SqlIndex, and_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


class DictMixin:
    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


Base = declarative_base(cls=DictMixin)


class Index(Base):
    __tablename__ = 'indexes'
    index_path = Column(String, primary_key=True, nullable=False)
    prefix = Column(String, nullable=False)
    org_id = Column(String)
    timestamp = Column(DateTime(timezone=True))
    size_MB = Column(Integer)
    s3_event_sequencer = Column(String, nullable=False)
    last_search_time = Column(DateTime(timezone=True))
    date_removed = Column(DateTime(timezone=True))
    priority = Column(Integer, server_default='0', nullable=False)
    partition = Column(String, nullable=True)
    retention_class = Column(String, nullable=True)
    last_modified = Column(DateTime(timezone=True))
    expiration_date = Column(DateTime(timezone=True))

    lease = relationship("Lease", cascade="all,delete", backref="leases")

    __table_args__ = (SqlIndex('indexes_date_removed_idx', date_removed),
                      SqlIndex('indexes_org_partition', org_id, partition),
                      SqlIndex('indexes_expiration_date_idx', expiration_date,
                               postgresql_where=(and_(expiration_date.isnot(None), date_removed.is_(None)))),)


class Lease(Base):
    __tablename__ = 'leases'
    index_path = Column(String, ForeignKey('indexes.index_path', ondelete='CASCADE'), primary_key=True, nullable=False)
    container_version = Column(String, nullable=False)
    forced = Column(Boolean)
    leased_to = Column(String)
    expiration = Column(DateTime(timezone=True))

    __table_args__ = (ForeignKeyConstraint((index_path, ), (Index.index_path, ), ondelete='CASCADE'),
                      {})


class Searcher(Base):
    __tablename__ = 'searchers'
    searcher_id = Column(String, primary_key=True, nullable=False)
    org_id = Column(String, primary_key=True, nullable=False)
    container_version = Column(String, nullable=False)
    endpoint = Column(String, nullable=False)
    starttime = Column(DateTime(timezone=True))
    endtime = Column(DateTime(timezone=True))
    index_path = Column(String, ForeignKey('indexes.index_path', ondelete='CASCADE'))
    index_partition_id = Column(String)
    index_stage = Column(Integer)
    expiration = Column(DateTime(timezone=True))

    __table_args__ = (ForeignKeyConstraint((index_path, ), (Index.index_path, ), ondelete='CASCADE'),
                      SqlIndex('searchers_expiration_idx', expiration),
                      SqlIndex('searchers_index_path_idx', index_path),)
